using System;
using System.Collections;
using System.Globalization;
using Microsoft.SqlServer.Dts.Runtime;
using Microsoft.SqlServer.Dts.Pipeline;
using Microsoft.SqlServer.Dts.Pipeline.Wrapper;
using Microsoft.SqlServer.Dts.Runtime.Wrapper;
using System.Data.SqlClient;
using System.Data;

namespace Bimp.SSIS
{
    [
        DtsPipelineComponent
        (
            DisplayName = "VRC Row Count"
            , Description = "Counts rows of different types (Select, Insert, Update, Delete, Intermediate, Error, Truncate), and logs this via the Bimp custom log provider."
            , ComponentType = 0
            , IconResource = "Bimp.SSIS.RowCount.ico"
        )]

    public class RowCount : PipelineComponent
    {
        // Fields
        SqlConnection _connection;
        private SqlCommand _cmd;
        private bool _pbCancel;
        private int _allRowCount;
        private decimal _allAggregate;
        string _inputColumnName;
        readonly ArrayList _mColumnIndexList = new ArrayList();
        
        // Methods
        public override void DeleteOutput(int outputID)
        {
            throw new Exception("Can't delete output " + outputID);
        }

        public override void OnInputPathAttached(int inputID)
        {
        }


        public override void PerformUpgrade(int pipelineVersion)
        {
            var num = ((DtsPipelineComponentAttribute)Attribute.GetCustomAttribute(GetType(), typeof(DtsPipelineComponentAttribute), false)).CurrentVersion;
            ComponentMetaData.Version = num;
        }

        public override void PostExecute()
        {
            var rowCountData = new RowCountSerialize
                {
                    RowCountType =
                        (RowCountSerialize.RowCountTypeEnum)
                        (ComponentMetaData.CustomPropertyCollection["RowCountType"].Value),
                    RowCount = _allRowCount
                };
            if (_inputColumnName == null)
            {
                rowCountData.ColumnSum = 0;
                rowCountData.ColumnName = "";

            }
            else
            {
                rowCountData.ColumnSum = _allAggregate;
                rowCountData.ColumnName = _inputColumnName;
            }
            rowCountData.RowCountComponent = ComponentMetaData.Name;
            rowCountData.RowCountObject = (string)(ComponentMetaData.CustomPropertyCollection["RowCountObject"].Value);

            VariableDispenser.LockForRead("User::ExecutionID");
            IDTSVariables100 variables;
            VariableDispenser.GetVariables(out variables);

            if (variables.Locked)
            {
                rowCountData.ExecutionID = (Int32)variables["User::ExecutionID"].Value;
                variables.Unlock();
            }

            _cmd.Parameters["@ExecutionID"].Value = rowCountData.ExecutionID;
            _cmd.Parameters["@ComponentName"].Value = rowCountData.RowCountComponent;
            _cmd.Parameters["@ObjectName"].Value = rowCountData.RowCountObject;
            _cmd.Parameters["@CountType"].Value = rowCountData.RowCountTypeString;
            _cmd.Parameters["@RowCount"].Value = rowCountData.RowCount;
            _cmd.Parameters["@ColumnSum"].Value = rowCountData.ColumnSum;
            _cmd.Parameters["@ColumnName"].Value = rowCountData.ColumnName;
            _cmd.ExecuteNonQuery();
        }

        public override void PreExecute()
        {
            var input = ComponentMetaData.InputCollection[0];
            var inputColumns = input.InputColumnCollection;

            _cmd = new SqlCommand("ssis.LogRowCount", _connection) {CommandType = CommandType.StoredProcedure};
            SqlCommandBuilder.DeriveParameters(_cmd); // Creates all command parameters from the SP parameters

            foreach (IDTSInputColumn100 column in inputColumns)
            {
                _mColumnIndexList.Add(BufferManager.FindColumnByLineageID(input.Buffer, column.LineageID));
                _inputColumnName = column.Name;
            }
            _allRowCount = 0;
            _allAggregate = 0;
        }

        public override void ProcessInput(int inputID, PipelineBuffer buffer)
        {
            if (buffer.EndOfRowset) return;
            while (buffer.NextRow())
            {

                foreach (int columnIndex in _mColumnIndexList)
                {
                    _allAggregate += decimal.Parse(buffer[columnIndex].ToString());
                }

            }
            _allRowCount += buffer.RowCount;
        }

        public override void ProvideComponentProperties()
        {
            RemoveAllInputsOutputsAndCustomProperties();
            ComponentMetaData.UsesDispositions = false;

            var input = ComponentMetaData.InputCollection.New();
            input.Name = "RowCountInput";
            input.HasSideEffects  = true;

            var output = ComponentMetaData.OutputCollection.New();
            output.Name = "RowCountOutput";
            output.SynchronousInputID = input.ID;
            var property = ComponentMetaData.CustomPropertyCollection.New();
            property.Name = "RowCountObject";
            property.Description = "An string that identifies what the row count refers to, e.g. a fully qualified table name, a file name or an arbitrary name for an intermediate processing step. This property can be set using expressions.";
            property.ExpressionType = DTSCustomPropertyExpressionType.CPET_NOTIFY;
            var property2 = ComponentMetaData.CustomPropertyCollection.New();
            property2.Name = "RowCountType";
            property2.Description  = "The type of row count, e.g. INSERT, UPDATE etc.";
            property2.TypeConverter = typeof(RowCountSerialize.RowCountTypeEnum).AssemblyQualifiedName;
            property2.Value = RowCountSerialize.RowCountTypeEnum.Unknown;

            var connection = ComponentMetaData.RuntimeConnectionCollection.New();
            connection.Name = "BIMP";

        }

        public override void AcquireConnections(object transaction)
        {
            if (ComponentMetaData.RuntimeConnectionCollection.Count <= 0) return;
            if (ComponentMetaData.RuntimeConnectionCollection[0].ConnectionManager == null) return;
            var cm = DtsConvert.GetWrapper(ComponentMetaData.RuntimeConnectionCollection[0].ConnectionManager);
            var bimp = cm.InnerObject as ConnectionManagerAdoNet;

            if (bimp == null)
                throw new Exception("The ConnectionManager " + cm.Name + " is not an ADO.NET connection.");

            _connection = bimp.AcquireConnection(transaction) as SqlConnection;
            if (_connection == null || _connection.State == ConnectionState.Open) return;
            _connection.Open();
        }

        public override void ReleaseConnections()
        {
            if (_connection != null && _connection.State != ConnectionState.Closed)
            {
                _connection.Close();
            }
        }

        public override IDTSInputColumn100 SetUsageType(int inputID, IDTSVirtualInput100 virtualInput, int lineageID, DTSUsageType usageType)
        {
            if (usageType == DTSUsageType.UT_READWRITE)
            {
                throw new Exception("Columns must be READ ONLY");
            }
            return base.SetUsageType(inputID, virtualInput, lineageID, usageType);
        }

        public override DTSValidationStatus Validate()
        {

            if (ComponentMetaData.InputCollection.Count != 1)
            {
                ComponentMetaData.FireError(0, ComponentMetaData.Name, "Incorrect number of inputs. Only one input can be used.", "", 0, out _pbCancel);
                return DTSValidationStatus.VS_ISCORRUPT;
            }
            if ((ComponentMetaData.CustomPropertyCollection["RowCountObject"].Value == null) || (ComponentMetaData.CustomPropertyCollection["RowCountObject"].Value.ToString().Length == 0))
            {
                ComponentMetaData.FireError(0, "RowCountObject", "The RowCountObject property must be set.", "", 0, out _pbCancel);
                return DTSValidationStatus.VS_ISBROKEN;
            }
            //	If there is an input column that no longer exists in the Virtual input collection,
            // return needs new meta data. The designer will then call ReinitalizeMetadata which will clean up the input collection.
            if (ComponentMetaData.AreInputColumnsValid == false)
            {
                return DTSValidationStatus.VS_NEEDSNEWMETADATA;
            }

            // Check that only one Input column exist in the collection:
            if (ComponentMetaData.InputCollection[0].InputColumnCollection.Count > 1)
            {
                ComponentMetaData.FireError(0, "RowCountObject", "Currently only one input column can be selected. This column must be numeric and will be aggregated.", "", 0, out _pbCancel);
                return DTSValidationStatus.VS_ISBROKEN;
            }
            // Check each Input column in the collection:
            // Validate column datatype.
            
            if (ComponentMetaData.InputCollection[0].InputColumnCollection.Count == 1)
            {
                var column = ComponentMetaData.InputCollection[0].InputColumnCollection[0];
                if (column.DataType != DataType.DT_CY &&
                    column.DataType != DataType.DT_I1 &&
                    column.DataType != DataType.DT_I2 &&
                    column.DataType != DataType.DT_I4 &&
                    column.DataType != DataType.DT_I8 &&
                    column.DataType != DataType.DT_NUMERIC &&
                    column.DataType != DataType.DT_R4 &&
                    column.DataType != DataType.DT_R8 &&
                    column.DataType != DataType.DT_UI1 &&
                    column.DataType != DataType.DT_UI2 &&
                    column.DataType != DataType.DT_UI4 &&
                    column.DataType != DataType.DT_UI8)
                {
                    ComponentMetaData.FireError(0, "RowCountObject", ComponentMetaData.InputCollection[0].InputColumnCollection[0].Name.ToString(CultureInfo.InvariantCulture) + " is not a numeric column. Please select a column of numeric type that you would like to aggregate.", "", 0, out _pbCancel);
                    return DTSValidationStatus.VS_ISBROKEN;
                }
            }

            
            return base.Validate();
        }
    }


}
